@@ -1,6 +1,7 @@
 import tzlocal
 from datetime import datetime, timezone
 from typing import List, Dict, Union, NoReturn, Any, Tuple
+from enum import Enum
 from pathlib import Path
 import math
 import pandas as pd
@@ -10,7 +11,9 @@ from hurst import compute_Hc # compatible with pypy
 from ccxt.base.exchange import Exchange as CcxtExchange
 from ccxt import deribit
 
+from siglab_py.util.simple_math import bucket_series, bucketize_val
 from siglab_py.util.market_data_util import fix_column_types
+from siglab_py.constants import TrendDirection
 
 # Fibonacci
 MAGIC_FIB_LEVELS = [0, 0.236, 0.382, 0.5, 0.618, 0.786, 1.00, 1.618, 2.618, 3.618, 4.236]
@@ -51,41 +54,45 @@ def calculate_slope(
     pd_data[f"normalized_{slope_col_name}_idmin"] = normalized_slope_rolling.apply(lambda x : x.idxmin())
     pd_data[f"normalized_{slope_col_name}_idmax"] = normalized_slope_rolling.apply(lambda x : x.idxmax())
 
-def higherhighs(series: pd.Series) -> str:
-    unique_maxima = series.dropna()[series.dropna().diff().ne(0)]
+def trend_from_highs(series: np.ndarray) -> float:
+    valid_series = series[~np.isnan(series)]
+    unique_maxima = valid_series[np.concatenate(([True], np.diff(valid_series) != 0))]
     if len(unique_maxima) < 2:
-        return 'sideways'
-    first, last = unique_maxima.iloc[0], unique_maxima.iloc[-1]
+        return TrendDirection.UNDEFINED.value
+    first, last = unique_maxima[0], unique_maxima[-1]
     if first > last:
-        return 'lower_highs'
+        return TrendDirection.LOWER_HIGHS.value
     elif first < last:
-        return 'higher_highs'
+        return TrendDirection.HIGHER_HIGHS.value
     else:
-        return 'sideways'
+        return TrendDirection.SIDEWAYS.value
 
-def lowerlows(series: pd.Series) -> str:
-    unique_minima = series.dropna()[series.dropna().diff().ne(0)]
+def trend_from_lows(series: np.ndarray) -> float:
+    valid_series = series[~np.isnan(series)]
+    unique_minima = valid_series[np.concatenate(([True], np.diff(valid_series) != 0))]
     if len(unique_minima) < 2:
-        return 'sideways'
-    first, last = unique_minima.iloc[0], unique_minima.iloc[-1]
+        return TrendDirection.UNDEFINED.value
+    first, last = unique_minima[0], unique_minima[-1]
     if first > last:
-        return 'lower_lows'
+        return TrendDirection.LOWER_LOWS.value
     elif first < last:
-        return 'higher_lows'
+        return TrendDirection.HIGHER_LOWS.value
     else:
-        return 'sideways'
-
+        return TrendDirection.SIDEWAYS.value
+    
+    
 '''
 compute_candles_stats will calculate typical/basic technical indicators using in many trading strategies:
     a. Basic SMA/EMAs (And slopes)
-    b. ATR
-    c. Boillenger bands (Yes incorrect spelling sorry)
-    d. FVG
-    e. Hurst Exponent
-    f. RSI, MFI
-    g. MACD
-    h. Fibonacci
-    i. Inflections points: where 'close' crosses EMA from above or below.
+    b. EMA crosses
+    c. ATR
+    d. Boillenger bands (Yes incorrect spelling sorry)
+    e. FVG
+    f. Hurst Exponent
+    g. RSI, MFI
+    h. MACD
+    i. Fibonacci
+    j. Inflections points: where 'close' crosses EMA from above or below.
 
 Parameters:
     a. boillenger_std_multiples: For boillenger upper and lower calc
@@ -115,6 +122,7 @@ def compute_candles_stats(
         pypy_compat : bool = True
         ):
     pd_candles['candle_height'] = pd_candles['high'] - pd_candles['low']
+    pd_candles['candle_body_height'] = pd_candles['close'] - pd_candles['open']
 
     '''
     market_data_gizmo inserted dummy lines --> Need exclude those or "TypeError: unorderable types for comparison": pd_btc_candles = pd_btc_candles[pd_btc_candles.close.notnull()]
@@ -136,7 +144,6 @@ def compute_candles_stats(
     close_short_periods_ewm = pd_candles['close'].ewm(span=int(sliding_window_how_many_candles/slow_fast_interval_ratio), adjust=False)
     close_long_periods_ewm = pd_candles['close'].ewm(span=sliding_window_how_many_candles, adjust=False)
 
-
     pd_candles['pct_change_close'] = pd_candles['close'].pct_change() * 100
     pd_candles['sma_short_periods'] = close_short_periods_rolling.mean()
     pd_candles['sma_long_periods'] = close_long_periods_rolling.mean()
@@ -152,6 +159,9 @@ def compute_candles_stats(
     pd_candles['candle_height_percent'] = pd_candles['candle_height'] / pd_candles['ema_close'] * 100
     pd_candles['candle_height_percent_rounded'] = pd_candles['candle_height_percent'].round().astype('Int64')
 
+    pd_candles['candle_body_height_percent'] = pd_candles['candle_body_height'] / pd_candles['ema_close'] * 100
+    pd_candles['candle_body_height_percent_rounded'] = pd_candles['candle_body_height_percent'].round().astype('Int64')
+
     '''
     To annualize volatility:
         if candle_interval == '1m':
@@ -163,6 +173,8 @@ def compute_candles_stats(
         pd_candles['annualized_volatility'] = (
             pd_candles['interval_historical_volatility'] * annualization_factor
         )
+
+    Why log return? Trading Dude https://python.plainenglish.io/stop-using-percentage-returns-logarithmic-returns-explained-with-code-64a4634b883a
     '''
     pd_candles['log_return'] = np.log(pd_candles['close'] / pd_candles['close'].shift(1))
     pd_candles['interval_hist_vol'] = pd_candles['log_return'].rolling(window=sliding_window_how_many_candles).std()
@@ -173,7 +185,6 @@ def compute_candles_stats(
     annualization_factor = np.sqrt(candles_per_year)
     pd_candles['annualized_hist_vol'] = pd_candles['interval_hist_vol'] * annualization_factor
 
-
     pd_candles['chop_against_ema'] = (
         (~pd_candles['is_green'] & (pd_candles['close'] > pd_candles['ema_close'])) |  # Case 1: Green candle and close > EMA
         (pd_candles['is_green'] & (pd_candles['close'] < pd_candles['ema_close']))   # Case 2: Red candle and close < EMA
@@ -182,6 +193,42 @@ def compute_candles_stats(
     pd_candles['ema_volume_short_periods'] = pd_candles['volume'].ewm(span=sliding_window_how_many_candles/slow_fast_interval_ratio, adjust=False).mean()
     pd_candles['ema_volume_long_periods'] = pd_candles['volume'].ewm(span=sliding_window_how_many_candles, adjust=False).mean()
 
+    pd_candles['ema_cross'] = None
+    pd_candles['ema_cross_last'] = None
+    pd_candles['ema_bullish_cross_last_id'] = None
+    pd_candles['ema_bearish_cross_last_id'] = None
+    ema_short_periods_prev = pd_candles['ema_short_periods'].shift(1)
+    ema_long_periods_prev = pd_candles['ema_long_periods'].shift(1)
+    ema_short_periods_curr = pd_candles['ema_short_periods']
+    ema_long_periods_curr = pd_candles['ema_long_periods']
+    bullish_ema_crosses = (ema_short_periods_prev <= ema_long_periods_prev) & (ema_short_periods_curr > ema_long_periods_curr)
+    bearish_ema_crosses = (ema_short_periods_prev >= ema_long_periods_prev) & (ema_short_periods_curr < ema_long_periods_curr)
+    pd_candles.loc[bullish_ema_crosses, 'ema_cross'] = 1
+    pd_candles.loc[bearish_ema_crosses, 'ema_cross'] = -1
+    bullish_indices = pd.Series(pd_candles.index.where(pd_candles['ema_cross'] == 1), index=pd_candles.index).astype('Int64')
+    bearish_indices = pd.Series(pd_candles.index.where(pd_candles['ema_cross'] == -1), index=pd_candles.index).astype('Int64')
+    pd_candles['ema_bullish_cross_last_id'] = bullish_indices.rolling(window=pd_candles.shape[0], min_periods=1).max().astype('Int64')
+    pd_candles['ema_bearish_cross_last_id'] = bearish_indices.rolling(window=pd_candles.shape[0], min_periods=1).max().astype('Int64')
+    conditions = [
+        (pd_candles['ema_bullish_cross_last_id'].notna() & 
+        pd_candles['ema_bearish_cross_last_id'].notna() &
+        (pd_candles['ema_bullish_cross_last_id'] > pd_candles['ema_bearish_cross_last_id'])),
+        
+        (pd_candles['ema_bullish_cross_last_id'].notna() & 
+        pd_candles['ema_bearish_cross_last_id'].notna() &
+        (pd_candles['ema_bearish_cross_last_id'] > pd_candles['ema_bullish_cross_last_id'])),
+        
+        (pd_candles['ema_bullish_cross_last_id'].notna() & 
+        pd_candles['ema_bearish_cross_last_id'].isna()),
+        
+        (pd_candles['ema_bearish_cross_last_id'].notna() & 
+        pd_candles['ema_bullish_cross_last_id'].isna())
+    ]
+    choices = ['bullish', 'bearish', 'bullish', 'bearish']
+    pd_candles['ema_cross_last'] = np.select(conditions, choices, default=None) # type: ignore
+    pd_candles.loc[bullish_ema_crosses, 'ema_cross'] = 'bullish'
+    pd_candles.loc[bearish_ema_crosses, 'ema_cross'] = 'bearish'
+
     pd_candles['max_short_periods'] = close_short_periods_rolling.max()
     pd_candles['max_long_periods'] = close_long_periods_rolling.max()
     pd_candles['idmax_short_periods'] = close_short_periods_rolling.apply(lambda x : x.idxmax())
@@ -192,6 +239,11 @@ def compute_candles_stats(
     pd_candles['idmin_short_periods'] = close_short_periods_rolling.apply(lambda x : x.idxmin())
     pd_candles['idmin_long_periods'] = close_long_periods_rolling.apply(lambda x : x.idxmin())
 
+    pd_candles['max_candle_body_height_percent_long_periods'] = pd_candles['candle_body_height_percent'].rolling(window=sliding_window_how_many_candles).max()
+    pd_candles['idmax_candle_body_height_percent_long_periods'] = pd_candles['candle_body_height_percent'].rolling(window=sliding_window_how_many_candles).apply(lambda x : x.idxmax())
+    pd_candles['min_candle_body_height_percent_long_periods'] = pd_candles['candle_body_height_percent'].rolling(window=sliding_window_how_many_candles).min()
+    pd_candles['idmin_candle_body_height_percent_long_periods'] = pd_candles['candle_body_height_percent'].rolling(window=sliding_window_how_many_candles).apply(lambda x : x.idxmin())
+
     pd_candles['price_swing_short_periods'] = np.where(
         pd_candles['idmax_short_periods'] > pd_candles['idmin_short_periods'],
         pd_candles['max_short_periods'] - pd_candles['min_short_periods'],  # Up swing
@@ -204,10 +256,26 @@ def compute_candles_stats(
         pd_candles['min_long_periods'] - pd_candles['max_long_periods']   # Down swing (negative)
     )
 
-    pd_candles['higher_highs_long_periods'] = higherhighs(pd_candles['max_long_periods'])
-    pd_candles['lower_lows_long_periods'] = lowerlows(pd_candles['min_long_periods'])
-    pd_candles['higher_highs_short_periods'] = higherhighs(pd_candles['max_short_periods'])
-    pd_candles['lower_lows_short_periods'] = lowerlows(pd_candles['min_short_periods'])
+    pd_candles['trend_from_highs_long_periods'] = np.where(
+												pd.isna(pd_candles['max_long_periods']),
+                                                None, # type: ignore
+												pd_candles['max_long_periods'].rolling(window=sliding_window_how_many_candles).apply(trend_from_highs, raw=True)
+												)
+    pd_candles['trend_from_lows_long_periods'] = np.where(
+												pd.isna(pd_candles['min_long_periods']),
+                                                None, # type: ignore
+												pd_candles['min_long_periods'].rolling(window=sliding_window_how_many_candles).apply(trend_from_lows, raw=True)
+												)
+    pd_candles['trend_from_highs_short_periods'] = np.where(
+												pd.isna(pd_candles['max_short_periods']),
+                                                None, # type: ignore
+												pd_candles['max_short_periods'].rolling(window=int(sliding_window_how_many_candles/slow_fast_interval_ratio)).apply(trend_from_highs, raw=True)
+												)
+    pd_candles['trend_from_lows_short_periods'] = np.where(
+												pd.isna(pd_candles['min_short_periods']),
+                                                None, # type: ignore
+												pd_candles['min_short_periods'].rolling(window=int(sliding_window_how_many_candles/slow_fast_interval_ratio)).apply(trend_from_lows, raw=True)
+												)
 
     # ATR https://medium.com/codex/detecting-ranging-and-trending-markets-with-choppiness-index-in-python-1942e6450b58 
     pd_candles.loc[:,'h_l'] = pd_candles['high'] - pd_candles['low']
@@ -375,8 +443,11 @@ def compute_candles_stats(
 
     pd_candles['fvg_mitigated'] = pd_candles.apply(lambda row: compute_fvg_mitigated(row, pd_candles), axis=1)
 
-
-    # RSI - https://www.youtube.com/watch?v=G9oUTi-PI18&t=809s 
+    '''
+        RSI
+        Divergences from Bybit Learn https://www.youtube.com/watch?v=G9oUTi-PI18&t=809s 
+        RSI Reversals from BK Traders https://www.youtube.com/watch?v=MvkbrHjiQlI
+    '''
     pd_candles.loc[:,'close_delta'] = pd_candles['close'].diff()
     pd_candles.loc[:,'close_delta_percent'] = pd_candles['close'].pct_change()
     lo_up = pd_candles['close_delta'].clip(lower=0)
@@ -402,6 +473,13 @@ def compute_candles_stats(
         
     lo_rs = lo_ma_up / lo_ma_down
     pd_candles.loc[:,'rsi'] = 100 - (100/(1 + lo_rs))
+    rsi_buckets = bucket_series(
+        values = pd_candles['rsi'].to_list(),
+        outlier_threshold_percent=10,
+        level_granularity=0.1
+    )
+    
+    pd_candles['rsi_bucket'] = pd_candles['rsi'].apply(lambda x: bucketize_val(x, buckets=rsi_buckets))
     pd_candles['ema_rsi'] = pd_candles['rsi'].ewm(
         span=rsi_sliding_window_how_many_candles, 
         adjust=False).mean()
@@ -425,18 +503,26 @@ def compute_candles_stats(
 
     pd_candles['rsi_trend'] = pd_candles.apply(lambda row: rsi_trend(row), axis=1)
 
-    pd_candles['rsi_higher_highs'] = higherhighs(pd_candles['rsi_max'])
-    pd_candles['rsi_lower_lows'] = lowerlows(pd_candles['rsi_min'])
+    pd_candles['rsi_trend_from_highs'] = np.where(
+												pd.isna(pd_candles['rsi_max']),
+                                                None, # type: ignore
+												pd_candles['rsi_max'].rolling(window=rsi_trend_sliding_window_how_many_candles).apply(trend_from_highs, raw=True)
+												)
+    pd_candles['rsi_trend_from_lows'] = np.where(
+												pd.isna(pd_candles['rsi_min']),
+                                                None, # type: ignore
+												pd_candles['rsi_min'].rolling(window=rsi_trend_sliding_window_how_many_candles).apply(trend_from_lows, raw=True)
+												)
 
     def _rsi_divergence(row):
-		# Bullish Divergence: Price lower low + RSI higher low
-        if row['lower_lows_long_periods']=='lower_lows' and row['rsi_lower_lows']=='higher_lows':
+        trend_from_highs_long_periods = TrendDirection(row['trend_from_highs_long_periods']) if row['trend_from_highs_long_periods'] is not None and not pd.isna(row['trend_from_highs_long_periods']) else None  # type: ignore
+        rsi_trend_from_highs = TrendDirection(row['rsi_trend_from_highs']) if row['rsi_trend_from_highs'] is not None and not pd.isna(row['rsi_trend_from_highs']) else None # type: ignore
+        
+        if trend_from_highs_long_periods and rsi_trend_from_highs and trend_from_highs_long_periods == TrendDirection.LOWER_HIGHS and rsi_trend_from_highs == TrendDirection.HIGHER_HIGHS:
             return 'bullish_divergence'
-		# Bearish Divergence: Price higher high + RSI lower high
-        elif row['higher_highs_long_periods']=='higher_highs' and row['rsi_higher_highs']=='lower_highs':
+        elif trend_from_highs_long_periods and rsi_trend_from_highs and trend_from_highs_long_periods == TrendDirection.HIGHER_HIGHS and rsi_trend_from_highs == TrendDirection.LOWER_HIGHS:
             return 'bearish_divergence'
-        else:
-            return 'no_divergence'
+        return 'no_divergence'
     pd_candles['rsi_divergence'] = pd_candles.apply(_rsi_divergence, axis=1)
     
 
@@ -455,6 +541,12 @@ def compute_candles_stats(
         rsi_sliding_window_how_many_candles if rsi_sliding_window_how_many_candles else sliding_window_how_many_candles).sum()
     pd_candles['money_flow_ratio'] = pd_candles['positive_flow_sum'] / pd_candles['negative_flow_sum']
     pd_candles['mfi'] = 100 - (100 / (1 + pd_candles['money_flow_ratio']))
+    mfi_buckets = bucket_series(
+        values = pd_candles['mfi'].to_list(),
+        outlier_threshold_percent=10,
+        level_granularity=0.1
+    )
+    pd_candles['mfi_bucket'] = pd_candles['mfi'].apply(lambda x: bucketize_val(x, buckets=mfi_buckets))
     
 
     # MACD https://www.investopedia.com/terms/m/macd.asp
@@ -510,7 +602,7 @@ def compute_candles_stats(
             pd_data=pd_candles,
             src_col_name='ema_rsi',
             slope_col_name='ema_rsi_slope',
-            sliding_window_how_many_candles=int(sliding_window_how_many_candles)
+            sliding_window_how_many_candles=int(rsi_trend_sliding_window_how_many_candles)
         )
 
         pd_candles['regular_divergence'] = (
@@ -533,6 +625,8 @@ def compute_candles_stats(
 
     # Inflection points
     pd_candles['gap_close_vs_ema'] = pd_candles['close'] - pd_candles['ema_long_periods']
+    pd_candles['gap_close_vs_ema_percent'] = pd_candles['gap_close_vs_ema']/pd_candles['close'] *100
+
     pd_candles['close_above_or_below_ema'] = None
     pd_candles.loc[pd_candles['gap_close_vs_ema'] > 0, 'close_above_or_below_ema'] = 'above'
     pd_candles.loc[pd_candles['gap_close_vs_ema'] < 0, 'close_above_or_below_ema'] = 'below'
@@ -542,6 +636,68 @@ def compute_candles_stats(
         'close_vs_ema_inflection'
     ] = np.sign(pd_candles['close'] - pd_candles['ema_long_periods'])
 
+def lookup_fib_target(
+            row,
+            pd_candles,
+            target_fib_level : float = 0.618
+        ) -> Union[Dict, None]:
+            if row is None:
+                return None
+            
+            fib_target_short_periods = None
+            fib_target_long_periods = None
+
+            max_short_periods = row['max_short_periods']
+            idmax_short_periods = int(row['idmax_short_periods']) if not math.isnan(row['idmax_short_periods']) else None
+            max_long_periods = row['max_long_periods']
+            idmax_long_periods = int(row['idmax_long_periods']) if not math.isnan(row['idmax_long_periods']) else None
+
+            min_short_periods = row['min_short_periods']
+            idmin_short_periods = int(row['idmin_short_periods']) if not math.isnan(row['idmin_short_periods']) else None
+            min_long_periods = row['min_long_periods']
+            idmin_long_periods = int(row['idmin_long_periods']) if not math.isnan(row['idmin_long_periods']) else None
+                        
+            if idmax_short_periods and idmin_short_periods and idmax_short_periods>0 and idmin_short_periods>0:
+                if idmax_short_periods>idmin_short_periods and idmax_short_periods < len(pd_candles):
+                    # Falling from prev peak
+                    last_peak = pd_candles.iloc[idmax_short_periods]
+                    fib_target_short_periods = last_peak[f'fib_{target_fib_level}_short_periods'] if not math.isnan(last_peak[f'fib_{target_fib_level}_short_periods']) else None
+                    
+                else:
+                    # Bouncing from prev bottom
+                    if idmin_short_periods < len(pd_candles):
+                        last_bottom = pd_candles.iloc[idmin_short_periods]
+                        fib_target_short_periods = last_bottom[f'fib_{target_fib_level}_short_periods'] if not math.isnan(last_bottom[f'fib_{target_fib_level}_short_periods']) else None
+                        
+            if idmax_long_periods and idmin_long_periods and idmax_long_periods>0 and idmin_long_periods>0:
+                if idmax_long_periods>idmin_long_periods and idmax_long_periods < len(pd_candles):
+                    # Falling from prev peak
+                    last_peak = pd_candles.iloc[idmax_long_periods]
+                    fib_target_long_periods = last_peak[f'fib_{target_fib_level}_long_periods'] if not math.isnan(last_peak[f'fib_{target_fib_level}_long_periods']) else None
+                    
+                else:
+                    # Bouncing from prev bottom
+                    if idmin_long_periods < len(pd_candles):
+                        last_bottom = pd_candles.iloc[idmin_long_periods]
+                        fib_target_long_periods = last_bottom[f'fib_{target_fib_level}_long_periods'] if not math.isnan(last_bottom[f'fib_{target_fib_level}_long_periods']) else None
+            
+            return {
+                'short_periods' : {
+                    'idmin' : idmin_short_periods,
+                    'idmax' : idmax_short_periods,
+                    'min' : min_short_periods,
+                    'max' : max_short_periods,
+                    'fib_target' : fib_target_short_periods,
+                },
+                'long_periods' : {
+                    'idmin' : idmin_long_periods,
+                    'idmax' : idmax_long_periods,
+                    'min' : min_long_periods,
+                    'max' : max_long_periods,
+                    'fib_target' : fib_target_long_periods
+                }
+            }
+
 ''' 
 The implementation from Geeksforgeeks https://www.geeksforgeeks.org/find-indices-of-all-local-maxima-and-local-minima-in-an-array/ is wrong. 
 If you have consecutive-duplicates, things will gall apart!
